from ..base import KitchenAITask, KitchenAITaskHookMixin
import functools
from ..schema import DependencyType

class StorageTask(KitchenAITask, KitchenAITaskHookMixin):
    """
    This is a class for registering storage tasks.
    """
    def __init__(self, namespace: str, dependency_manager=None):
        super().__init__(namespace, dependency_manager)
        self.namespace = namespace

    def handler(self, label: str, *dependencies: DependencyType):
        """Decorator for registering storage tasks with dependencies."""
        def decorator(func):
            @functools.wraps(func)
            @self.with_dependencies(*dependencies)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return self.register_task(label, wrapper)
        return decorator

    def on_store(self, label: str, *dependencies: DependencyType):
        """Decorator for registering storage hooks with dependencies."""
        def decorator(func):
            @self.with_dependencies(*dependencies)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return self.register_hook(label, "on_store", wrapper)
        return decorator
    
    def on_delete(self, label: str, *dependencies: DependencyType):
        """Decorator for registering deletion hooks with dependencies."""
        def decorator(func):
            @self.with_dependencies(*dependencies)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return self.register_hook(label, "on_delete", wrapper)
        return decorator

    def on_retrieve(self, label: str, *dependencies: DependencyType):
        """Decorator for registering retrieval hooks with dependencies."""
        def decorator(func):
            @self.with_dependencies(*dependencies)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return self.register_hook(label, "on_retrieve", wrapper)
        return decorator





from whisk.kitchenai_sdk.kitchenai import KitchenAIApp
from whisk.kitchenai_sdk.schema import (
    WhiskQuerySchema,
    WhiskQueryBaseResponseSchema,
    WhiskStorageSchema,
    WhiskStorageResponseSchema,
    WhiskEmbedSchema,
    WhiskEmbedResponseSchema,
)

try:
    from llama_index.llms.openai import OpenAI
except ImportError:
    raise ImportError("Please install llama-index to use this example")

import logging

# Initialize LLM and embeddings
llm = OpenAI(model="gpt-3.5-turbo")

kitchen = KitchenAIApp(namespace="example_bento_box")

# pip install llama-index
logger = logging.getLogger(__name__)


@kitchen.query.handler("query")
async def query_handler(data: WhiskQuerySchema) -> WhiskQueryBaseResponseSchema:
    """Query handler"""

    response = await llm.acomplete(data.query)

    print(response)


    return WhiskQueryBaseResponseSchema.from_llm_invoke(
        data.query,
        response.text,
    )


@kitchen.query.handler("stream")
async def stream_handler(data: WhiskQuerySchema) -> WhiskQueryBaseResponseSchema:
    """Query handler"""

    completions = llm.astream_complete(data.query)

    async def stream_generator():
        async for completion in completions:
            yield WhiskQueryBaseResponseSchema.from_llm_invoke(
                data.query,
                completion.delta,
            )

    return WhiskQueryBaseResponseSchema(
        input=data.query,
        stream_gen=stream_generator,
    )


@kitchen.storage.handler("storage")
async def storage_handler(data: WhiskStorageSchema) -> WhiskStorageResponseSchema:
    """Storage handler"""
    print("storage handler")

    return WhiskStorageResponseSchema(
        id=data.id,
        data=data.data,
        metadata=data.metadata,
    )

@kitchen.storage.on_delete("storage")
async def storage_delete_handler(data: WhiskStorageSchema) -> None:
    """Storage delete handler"""
    print("storage delete handler")
    print(data)


@kitchen.embeddings.handler("embed")
async def embed_handler(data: WhiskEmbedSchema) -> WhiskEmbedResponseSchema:
    """Embed handler"""
    print("embed handler")
    print(data)
    return WhiskEmbedResponseSchema(
        text=data.text,
        metadata=data.metadata,
    )

@kitchen.embeddings.on_delete("embed")
async def embed_delete_handler(data: WhiskEmbedSchema) -> None:
    """Embed delete handler"""
    print("embed delete handler")
    print(data)



from pydantic import BaseModel, ConfigDict, computed_field
from typing import List, Optional, Dict, Any, Callable
from enum import StrEnum, auto


class TokenCountSchema(BaseModel):
    embedding_tokens: Optional[int] = None
    llm_prompt_tokens: Optional[int] = None 
    llm_completion_tokens: Optional[int] = None
    total_llm_tokens: Optional[int] = None


class WhiskQuerySchema(BaseModel):
    query: str
    stream: bool = False
    stream_id: Optional[str] = None
    metadata: Optional[Dict[str, str]] = None
    label: Optional[str]

    # OpenAI Chat Completion Schema as optional for more context. Will come is as a dict.
    messages: Optional[List[object]] = None



class SourceNodeSchema(BaseModel):
    text: str
    metadata: Dict
    score: float

class WhiskQueryBaseResponseSchema(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    
    input: Optional[str] = None
    output: Optional[str] = None
    retrieval_context: Optional[List[SourceNodeSchema]] = None
    stream_gen: Any | None = None
    metadata: Optional[Dict[str, Any]] = {}
    token_counts: Optional[TokenCountSchema] = None


    # OpenAI Chat Completion Schema as optional for more context. Will come is as a dict.
    messages: Optional[List[object]] = None
    
    @classmethod
    def from_llama_response(cls, data, response, metadata=None, token_counts: TokenCountSchema | None = None):
        source_nodes = []
        if hasattr(response, 'source_nodes'):
            for node in response.source_nodes:
                source_nodes.append(SourceNodeSchema(
                    text=node.node.text,
                    metadata=node.node.metadata,
                    score=node.score
                ))
        if metadata and response.metadata:
            response.metadata.update(metadata)
        return cls(
            input=data.query,
            output=response.response,
            retrieval_context=source_nodes,
            metadata=response.metadata,
            token_counts=token_counts
        )
    
    @classmethod
    def from_llama_response_stream(cls, data, response, stream_gen, metadata: dict[str, Any] | None = {}, token_counts: TokenCountSchema | None = None):
        source_nodes = []
        if hasattr(response, 'source_nodes'):
            for node in response.source_nodes:
                source_nodes.append(SourceNodeSchema(
                    text=node.node.text,
                    metadata=node.node.metadata,
                    score=node.score
                ))

        if metadata:
            response.metadata.update(metadata)
        return cls(
            input=data.query,
            retrieval_context=source_nodes,
            metadata=response.metadata,
            stream_gen=stream_gen,
            token_counts=token_counts
        )
    
    @classmethod
    def with_string_retrieval_context(cls, data, response: str, retrieval_context: List[str], metadata: dict[str, Any] | None = {}, token_counts: TokenCountSchema | None = None):
        return cls(
            input=data.query,
            output=response.response,
            retrieval_context=[SourceNodeSchema(text=context, metadata=metadata, score=1.0) for context in retrieval_context],
            metadata=response.metadata,
            token_counts=token_counts
        )
    
    @classmethod
    def from_llm_invoke(cls, input: str, output: str, metadata=None, token_counts: TokenCountSchema | None = None):        
        return cls(
            input=input,
            output=output,
            metadata=metadata,
            token_counts=token_counts
        )

class WhiskStorageStatus(StrEnum):
    PENDING = "pending"
    ERROR = "error"
    COMPLETE = "complete"
    ACK = "ack"

class WhiskStorageSchema(BaseModel):
    id: int
    name: str
    label: str 
    data: Optional[bytes] = bytes()
    metadata: Optional[Dict[str, str]] = None
    extension: Optional[str] = None

class WhiskStorageGetRequestSchema(BaseModel):
    id: int
    presigned: bool = False


class WhiskStorageGetResponseSchema(BaseModel):
    presigned_url: Optional[str] = None
    error: Optional[str] = None

class WhiskStorageResponseSchema(BaseModel):
    id: int
    status: WhiskStorageStatus = WhiskStorageStatus.PENDING
    error: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None
    token_counts: Optional[TokenCountSchema] = None

    @classmethod
    def with_token_counts(cls, token_counts: TokenCountSchema):
        return cls(token_counts=token_counts)

class WhiskAgentResponseSchema(BaseModel):  
    response: str

class WhiskEmbedSchema(BaseModel):
    label: str
    text: Optional[str] = None
    metadata: Optional[Dict[str, str]] = None

class WhiskEmbedResponseSchema(BaseModel):
    metadata: Optional[Dict[str, Any]] = None
    token_counts: Optional[TokenCountSchema] = None

    @classmethod
    def with_token_counts(cls, token_counts: TokenCountSchema):
        return cls(token_counts=token_counts)

class WhiskBroadcastSchema(BaseModel):
    """Schema for broadcast messages"""
    message: str
    type: str = "info"  # info, warning, error, etc.
    metadata: Optional[Dict[str, Any]] = None

class WhiskBroadcastResponseSchema(BaseModel):
    """Schema for broadcast responses"""
    message: str
    type: str
    metadata: Optional[Dict[str, Any]] = None
    token_counts: Optional[TokenCountSchema] = None

    @classmethod
    def from_broadcast(cls, broadcast: WhiskBroadcastSchema, token_counts: TokenCountSchema | None = None):
        return cls(
            message=broadcast.message,
            type=broadcast.type,
            metadata=broadcast.metadata,
            token_counts=token_counts
        )


class NatsMessageMetadata(BaseModel):
    content_type: str
    correlation_id: str
    reply_to: Optional[str] = None
    message_id: str

class NatsMessage(BaseModel):
    """
    Used for Request/Response messages
    """
    body: bytes
    headers: Dict[str, str]
    metadata: NatsMessageMetadata
    decoded_body: Dict[str, Any]

    @classmethod
    def from_faststream(cls, msg):
        return cls(
            body=msg.body,
            headers=msg.headers,
            metadata=NatsMessageMetadata(
                content_type=msg.content_type,
                correlation_id=msg.correlation_id,
                reply_to=msg.reply_to,
                message_id=msg.message_id,
                request_id=msg._decoded_body.get('request_id'),
            subject=msg.raw_message.subject,
            client_id=msg._decoded_body.get('client_id')
            ),
            decoded_body=msg._decoded_body
        )

class DependencyType(StrEnum):
    """Types of dependencies that can be managed."""
    LLM = auto()
    VECTOR_STORE = auto()
    EMBEDDING = auto()  # For future use
    RETRIEVER = auto()  # For future use
    PROMPT = auto()     # For future use 
    MEMORY = auto()     # For future use


from faststream import FastStream, Logger

from faststream.nats import NatsBroker, PullSub, JStream


from contextlib import asynccontextmanager
from whisk.kitchenai_sdk.kitchenai import KitchenAIApp
import time
import sys
from nats.errors import Error as NatsError
import logging
from whisk.kitchenai_sdk.nats_schema import (
    QueryRequestMessage,
    StorageRequestMessage,
    EmbedRequestMessage,
    QueryResponseMessage,
    StorageResponseMessage,
    EmbedResponseMessage,
    BroadcastRequestMessage,
    NatsRegisterMessage,
    StorageGetRequestMessage,
    StorageGetResponseMessage,
)
from rich.console import Console
from .kitchenai_sdk.schema import (
    WhiskQuerySchema,
    WhiskStorageSchema,
    WhiskEmbedSchema,
    NatsMessage,
    WhiskStorageStatus,
)
import httpx

console = Console()


class WhiskClientError(Exception):
    """Base exception for WhiskClient errors"""

    pass


class WhiskAuthError(WhiskClientError):
    """Authentication/Authorization errors"""

    pass


class WhiskConnectionError(WhiskClientError):
    """Connection-related errors"""

    pass


logger = logging.getLogger(__name__)


class WhiskClient:
    """
    # As a client
    client = WhiskClient(user="clienta", password="...")
    await client.query("What is the temperature?", metadata={"location": "kitchen"})

    # As the KitchenAI service
    kitchenai = WhiskClient(user="kitchenai_admin", password="...", is_kitchenai=True)
    """

    def __init__(
        self,
        nats_url: str = "nats://localhost:4222",
        client_id: str = "whisk_client",
        user: str = None,
        password: str = None,
        is_kitchenai: bool = False,
        kitchen: KitchenAIApp = None,
        app: FastStream = None,
    ):
        self.client_id = client_id
        self.user = user
        self.is_kitchenai = is_kitchenai
        self.kitchen = kitchen
        self.app = app
        try:
            self.broker = NatsBroker(
                nats_url, name=client_id, user=user, password=password
            )

            self.app = FastStream(
                broker=self.broker, title=f"Whisk-{client_id}", lifespan=self.lifespan
            )

            # Register subscribers immediately
            if not self.is_kitchenai:
                self._setup_subscribers()

        except NatsError as e:
            if "Authorization" in str(e):
                raise WhiskAuthError(
                    f"Authentication failed for user '{user}'. Please check credentials."
                ) from e
            else:
                raise WhiskConnectionError(
                    f"Failed to connect to NATS: {str(e)}"
                ) from e
        except Exception as e:
            raise WhiskClientError(f"Failed to initialize WhiskClient: {str(e)}") from e

    @asynccontextmanager
    async def lifespan(self):
        try:
            yield
        except NatsError as e:
            if "Authorization" in str(e):
                logger.error(f"Authorization error: {str(e)}")
                sys.exit(1)  # Exit gracefully on auth errors
            elif "permissions violation" in str(e).lower():
                logger.error(f"Permissions error: {str(e)}")
                # Continue running but log the error
            else:
                logger.error(f"NATS error: {str(e)}")
        except Exception as e:
            logger.error(f"Unexpected error: {str(e)}")
        finally:
            if hasattr(self, "broker"):
                await self.broker.close()

    async def register_client(
        self, message: NatsRegisterMessage
    ) -> NatsRegisterMessage:
        """Used by the workers to register with the server"""
        ack = await self.broker.request(
            message, f"kitchenai.service.{message.client_id}.mgmt.register"
        )
        return ack

    def _setup_subscribers(self):
        # Update topic pattern to include client name
        client_prefix = (
            f"kitchenai.service.{self.client_id}"
            if not self.is_kitchenai
            else "kitchenai.service"
        )


        args = ("queue",)
        # Setup subscribers
        self.handle_query = self.broker.subscriber(f"{client_prefix}.query.*", *args)(
            self._handle_query
        )
        self.handle_heartbeat = self.broker.subscriber(
            f"{client_prefix}.heartbeat", *args
        )(self._handle_heartbeat)

        # TODO: Uncomment this when we have streaming WIP
        # self.handle_query_stream = self.broker.subscriber(
        #     f"{client_prefix}.query.*.stream", *args
        # )(self._handle_query_stream)
        self.handle_storage = self.broker.subscriber(
            f"{client_prefix}.storage.*",
            *args,
        )(self._handle_storage)
        self.handle_storage_delete = self.broker.subscriber(
            f"{client_prefix}.storage.*.delete",
            *args,
        )(self._handle_storage_delete)

        # TODO: Uncomment this when we have embeddings WIP
        # self.handle_embed = self.broker.subscriber(
        #     f"{client_prefix}.embedding.*", *args
        # )(self._handle_embed)
        # self.handle_embed_delete = self.broker.subscriber(
        #     f"{client_prefix}.embedding.*.delete", *args
        # )(self._handle_embed_delete)

    async def _handle_query(
        self, msg: QueryRequestMessage, logger: Logger
    ) -> QueryResponseMessage:
        logger.info(f"Query request: {msg}")
        task = self.kitchen.query.get_task(msg.label)
        if not task:
            return QueryResponseMessage(
                request_id=msg.request_id,
                timestamp=time.time(),
                client_id=msg.client_id,
                label=msg.label,
                output=None,
                retrieval_context=None,
                stream_gen=None,
                metadata=None,
                token_counts=None,
                error=f"No task found for query",
                messages=msg.messages,
            )

        response = await task(WhiskQuerySchema(**msg.model_dump()))
        response_dict = response.model_dump()

        # Update metadata with additional fields
        metadata = response_dict.get("metadata", {}) or {}  # Handle None case
        metadata.update(msg.metadata)
        response_dict["metadata"] = metadata
        response_dict["messages"] = msg.messages

        query_response = QueryResponseMessage(
            **response_dict,
            label=msg.label,
            client_id=msg.client_id,
            request_id=msg.request_id,
            timestamp=time.time(),
        )
        return query_response

    async def _handle_query_stream(
        self, msg: QueryRequestMessage, logger: Logger
    ) -> None:
        logger.info(f"Query stream request: {msg}")
        resp = await self.kitchen.query.get_task("stream")(
            WhiskQuerySchema(**msg.model_dump())
        )
        async for chunk in resp.stream_gen():
            await self._publish_stream(
                QueryResponseMessage(
                    request_id=msg.request_id,
                    timestamp=time.time(),
                    client_id=msg.client_id,
                    label=msg.label,
                    output=chunk,
                    metadata=msg.metadata,
                )
            )

    async def _handle_heartbeat(self, msg: NatsRegisterMessage, logger: Logger) -> None:
        logger.info(f"Heartbeat request: {msg}")
        return NatsRegisterMessage(
            client_id=msg.client_id,
            version=msg.version,
            name=msg.name,
            ack=True,
            message="heartbeat",
        )

    async def _handle_storage(self, msg: StorageRequestMessage, logger: Logger) -> None:
        """
        This is a storage request.
        Flow:
        - KitchenAI will publish a message to client bento box
        - Bento box will pick up the message and request a presigned url from KitchenAI
        - KitchenAI will return a presigned url
        - Bento box will use the presigned url to download the file
        - Bento box will process the file with the storage task
        - Bento box service will send a response back to the client for progress
        - KitchenAI will update object status.
        """
        logger.info(f"Storage request: {msg}")
        # Get the task handler
        task = self.kitchen.storage.get_task(msg.label)
        if not task:
            payload = StorageResponseMessage(
                id=msg.id,
                request_id=msg.request_id,
                timestamp=time.time(),
                client_id=msg.client_id,
                label=msg.label,
                status=WhiskStorageStatus.ERROR,
                error="No task found for storage request",
            )
            logger.error(f"Error processing storage request: {payload}")
            await self.broker.publish(
                payload,
                f"kitchenai.service.{msg.client_id}.storage.{msg.label}.response",
            )
            return
        # Get file pre-signed url from kitchenai storage
        try:
            nats_response = await self.broker.request(
                StorageGetRequestMessage(
                    id=msg.id,
                    request_id=msg.request_id,
                    timestamp=time.time(),
                    label=msg.label,
                    client_id=msg.client_id,
                    presigned=True,
                ),
                f"kitchenai.service.{msg.client_id}.storage.{msg.label}.get",
            )
        except Exception as e:
            logger.error(f"Error getting presigned url: {e}")
            await self.broker.publish(
                StorageResponseMessage(
                    id=msg.id,
                    request_id=msg.request_id,
                    timestamp=time.time(),
                    error=str(e),
                    label=msg.label,
                    client_id=msg.client_id,
                    status=WhiskStorageStatus.ERROR,
                ),
                f"kitchenai.service.{msg.client_id}.storage.{msg.label}.response",
            )
            return
        presigned_message = StorageGetResponseMessage(
            id=msg.id, **NatsMessage.from_faststream(nats_response).decoded_body
        )
        if presigned_message.error:
            raise WhiskClientError(
                f"Error getting presigned url: {presigned_message.error}"
            )
        logger.info(f"Presigned url: {presigned_message.presigned_url}")
        # Use httpx to download the file using the presigned URL
        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(presigned_message.presigned_url)
                if response.status_code != 200:
                    raise WhiskClientError(
                        f"Error downloading file: {response.status_code}"
                    )
                file_data = response.content
        except Exception as e:
            logger.error(f"Error downloading file: {e}")
            await self.broker.publish(
                StorageResponseMessage(
                    id=msg.id,
                    request_id=msg.request_id,
                    timestamp=time.time(),
                    error=str(e),
                    label=msg.label,
                    client_id=msg.client_id,
                    status=WhiskStorageStatus.ERROR,
                ),
                f"kitchenai.service.{msg.client_id}.storage.{msg.label}.response",
            )
            return

        # Process file with kitchen task
        try:
            response = await task(
                WhiskStorageSchema(
                    id=msg.id,
                    name=msg.name,
                    label=msg.label,
                    data=file_data,
                    metadata=msg.metadata,
                )
            )
        except Exception as e:
            logger.error(f"Error processing storage request: {e}")
            await self.broker.publish(
                StorageResponseMessage(
                    id=msg.id,
                    request_id=msg.request_id,
                    timestamp=time.time(),
                    label=msg.label,
                    client_id=msg.client_id,
                    metadata=response.metadata,
                    status=WhiskStorageStatus.ERROR,
                    token_counts=response.token_counts,
                    error=str(e),
                ),
                f"kitchenai.service.{msg.client_id}.storage.{msg.label}.response",
            )
        await self.broker.publish(
            StorageResponseMessage(
                id=msg.id,
                request_id=msg.request_id,
                timestamp=time.time(),
                label=msg.label,
                client_id=msg.client_id,
                metadata=response.metadata,
                status=WhiskStorageStatus.COMPLETE,
                token_counts=response.token_counts,
            ),
            f"kitchenai.service.{msg.client_id}.storage.{msg.label}.response",
        )

    async def _handle_storage_delete(
        self, msg: StorageRequestMessage, logger: Logger
    ) -> None:
        logger.info(f"Storage delete request: {msg}")
        task = self.kitchen.storage.get_hook(msg.label, "on_delete")
        if not task:
            logger.error(f"No task found for storage delete request: {msg.label}")
            return
        await task(WhiskStorageSchema(**msg.model_dump()))

    async def _handle_embed(self, msg: EmbedRequestMessage, logger: Logger) -> None:
        """
        This is an embed request.
        This is an object stored in kitchenai but since its just a text payload, we can serialize and send it directly.
        """
        logger.info(f"Embed request: {msg}")
        try:
            # Get the task handler
            task = self.kitchen.embeddings.get_task(msg.label)
            if not task:
                embed_response = EmbedResponseMessage(
                    id=msg.id,
                    request_id=msg.request_id,
                    timestamp=time.time(),
                    label=msg.label,
                    client_id=msg.client_id,
                    error="No task found for embed request",
                )
                await self.broker.publish(
                    embed_response,
                    f"kitchenai.service.{msg.client_id}.embedding.{msg.label}.response",
                )
                return
            response = await task(WhiskEmbedSchema(**msg.model_dump()))
            await self.broker.publish(
                EmbedResponseMessage(
                    id=msg.id,
                    request_id=msg.request_id,
                    timestamp=time.time(),
                    label=msg.label,
                    client_id=msg.client_id,
                    **response.model_dump(),
                ),
                f"kitchenai.service.{msg.client_id}.embedding.{msg.label}.response",
            )
        except Exception as e:
            logger.error(f"Error processing embed request: {str(e)}")
            await self.broker.publish(
                EmbedResponseMessage(
                    id=msg.id,
                    request_id=msg.request_id,
                    timestamp=time.time(),
                    error=str(e),
                    label=msg.label,
                    client_id=msg.client_id,
                ),
                f"kitchenai.service.{msg.client_id}.embedding.{msg.label}.response",
            )

    async def _handle_embed_delete(
        self, msg: EmbedRequestMessage, logger: Logger
    ) -> None:
        logger.info(f"Embed delete request: {msg}")
        task = self.kitchen.embeddings.get_hook(msg.label, "on_delete")
        if not task:
            logger.error(f"No task found for embed delete request: {msg.label}")
            return
        await task(WhiskEmbedSchema(**msg.model_dump()))

    async def _publish_stream(self, message: QueryResponseMessage):
        await self.broker.publish(
            message,
            f"kitchenai.service.{message.client_id}.query.{message.label}.stream.response",
        )

    async def query(self, message: QueryRequestMessage) -> NatsMessage:
        """Send a query request.
        Returns a NatsMessage object
        """
        response = await self.broker.request(
            message,
            f"kitchenai.service.{message.client_id}.query.{message.label}",
            timeout=10,
        )
        return NatsMessage.from_faststream(response)

    async def query_stream(self, message: QueryRequestMessage):
        """Send a query stream request. This will only work for KitchenAI Server
        KitchenAI will be subscribed to kitchenai.service.*.query.*.stream.response
        and will publish to SSE clients
        """
        await self.broker.publish(
            message,
            f"kitchenai.service.{message.client_id}.query.{message.label}.stream",
        )

    async def register_client(self, client_id: str) -> NatsRegisterMessage:
        """Used by the workers to register with the server. Request/Reply always returns a nats message"""
        response = await self.broker.request(
            NatsRegisterMessage(
                client_id=client_id,
                version=self.kitchen.version,
                name=self.kitchen.namespace,
                bento_box=self.kitchen.to_dict(),
                client_type=self.kitchen.client_type,
                client_description=self.kitchen.client_description,
            ),
            f"kitchenai.service.{client_id}.mgmt.register",
        )
        return NatsMessage.from_faststream(response)

    async def store_message(self, message: StorageRequestMessage):
        """Send a storage request"""
        await self.broker.publish(
            message, f"kitchenai.service.{message.client_id}.storage.{message.label}"
        )

    async def store_delete(self, message: StorageRequestMessage):
        """Send a storage delete request"""
        await self.broker.publish(
            message,
            f"kitchenai.service.{message.client_id}.storage.{message.label}.delete",
        )

    async def embed(self, message: EmbedRequestMessage):
        """Send an embed request"""
        logger.info(f"Embedding request: {message}")
        await self.broker.publish(
            message, f"kitchenai.service.{message.client_id}.embedding.{message.label}"
        )

    async def embed_delete(self, message: EmbedRequestMessage):
        """Send an embed delete request"""
        await self.broker.publish(
            message,
            f"kitchenai.service.{message.client_id}.embedding.{message.label}.delete",
        )

    async def broadcast(self, message: BroadcastRequestMessage):
        """Send a broadcast message"""
        await self.broker.publish(message, f"kitchenai.broadcast.{message.label}")

    async def run(self):
        async with self.app.broker:
            response = await self.register_client(self.client_id)
            logger.info(f"Registration response: {response.decoded_body}")

            if response.decoded_body.get("error"):
                error_msg = response.decoded_body.get("error")
                logger.error(f"Registration failed: {error_msg}")
                console.print(f"[bold red]Registration Error: {error_msg}[/bold red]")
                raise Exception(error_msg)

            # Pretty print the bento box configuration
            if self.kitchen:
                console.print("\n[bold blue]KitchenAI Configuration:[/bold blue]")
                console.print(self.kitchen.to_dict(), style="cyan")

            console.print(
                "\n[bold green]✓ Successfully registered client![/bold green]"
            )

        await self.app.run()



# /// script
# requires-python = ">=3.11"
# dependencies = [
#   "kitchenai-whisk",
#   "llama-index",
# ]
# ///

from whisk.kitchenai_sdk.kitchenai import KitchenAIApp
from whisk.kitchenai_sdk.schema import (
    WhiskQuerySchema,
    WhiskQueryBaseResponseSchema,
    WhiskStorageSchema,
    WhiskStorageResponseSchema,
    WhiskEmbedSchema,
    WhiskEmbedResponseSchema,
)

from llama_index.llms.openai import OpenAI
import logging
import asyncio
from whisk.client import WhiskClient
# Initialize LLM and embeddings
llm = OpenAI(model="gpt-3.5-turbo")

kitchen = KitchenAIApp(namespace="example_query")

# pip install llama-index
logger = logging.getLogger(__name__)


@kitchen.query.handler("query")
async def query_handler(data: WhiskQuerySchema) -> WhiskQueryBaseResponseSchema:
    """Query handler"""

    response = await llm.acomplete(data.query)

    print(response)

    return WhiskQueryBaseResponseSchema.from_llm_invoke(
        data.query,
        response.text,
    )


@kitchen.storage.handler("storage")
async def storage_handler(data: WhiskStorageSchema) -> WhiskStorageResponseSchema:
    """Storage handler"""
    print("storage handler")

    return WhiskStorageResponseSchema(
        id=data.id,
        data=data.data,
        metadata=data.metadata,
    )

@kitchen.storage.on_delete("storage")
async def storage_delete_handler(data: WhiskStorageSchema) -> None:
    """Storage delete handler"""
    print("storage delete handler")
    print(data)


@kitchen.embeddings.handler("embed")
async def embed_handler(data: WhiskEmbedSchema) -> WhiskEmbedResponseSchema:
    """Embed handler"""
    print("embed handler")
    print(data)
    return WhiskEmbedResponseSchema(
        text=data.text,
        metadata=data.metadata,
    )

@kitchen.embeddings.on_delete("embed")
async def embed_delete_handler(data: WhiskEmbedSchema) -> None:
    """Embed delete handler"""
    print("embed delete handler")
    print(data)

if __name__ == "__main__":

    client = WhiskClient(
        nats_url="nats://localhost:4222",
        client_id="whisk_client",
        user="playground",
        password="kitchenai_playground",
        kitchen=kitchen,
    )
    async def start():
        await client.run()

    try:
        asyncio.run(start())
    except KeyboardInterrupt:
        print("\nShutting down gracefully...")



from pydantic import BaseModel  


"""
Extra body for file requests using the OpenAI API
"""

class FileExtraBody(BaseModel):
    client_id: str
    namespace: str
    label: str
    version: str | None = None
    metadata: str | None = None

class ChatExtraBody(BaseModel):
    namespace: str | None = None
    version: str | None = None



from typer.testing import CliRunner
from whisk.whisk.cli import app

runner = CliRunner()

def test_version():
    result = runner.invoke(app, ["--version"])
    assert result.exit_code == 0
    assert "KitchenAI Whisk v" in result.stdout 